import json

from replay.models import Action

def escape(text):
    "Escape text for string.Template substitution."
    return text.replace('$', '$$')


class RecorderMiddleware(object):
    def __init__(self, get_response):
        self.get_response = get_response


    def __call__(self, request):
        "Create Action object based on request and response."
        kwargs = {'indent': 4, 'separators': (',', ': ')}
        method = request.method
        data = json.dumps(getattr(request, method), **kwargs)
        files_names = {key: value.name for key, value in request.FILES.items()}
        files = json.dumps(files_names, **kwargs)
        response = self.get_response(request)
        status_code = response.status_code
        redirect = 300 <= status_code < 400
        response_content = response.content
        try:
            response_content.decode('utf-8')
        except UnicodeDecodeError:
            response_content = ''
        content = response_content if not redirect else response.url

        Action.objects.create(
            method=method,
            path=escape(request.path),
            data=escape(data),
            files=escape(files),
            status_code=str(status_code),
            content=content,
        )

        return response



import re
import time

from django import forms
from django.contrib import admin
from django.core.urlresolvers import reverse
from django.db.models import Count
from django.shortcuts import redirect
from django.utils.html import format_html, format_html_join
from django.utils.safestring import mark_safe

from replay.models import Action, Validator, Scenario

class ActionAdminForm(forms.ModelForm):
    class Meta:
        model = Action
        fields = []  # Required but ignored.
        widgets = {
            'name': forms.TextInput(attrs={'size': 50}),
            'path': forms.TextInput(attrs={'size': 50}),
            'data': forms.Textarea(attrs={'rows': 4, 'cols': 80}),
            'files': forms.Textarea(attrs={'rows': 2, 'cols': 80}),
            'content': forms.Textarea(attrs={'rows': 8, 'cols': 80}),
        }

class ValidatorInline(admin.StackedInline):
    model = Validator
    extra = 8
    ordering = (
        'order',
        'id',
    )
    fields = (
        'order',
        'pattern',
    )

class ActionAdmin(admin.ModelAdmin):
    def create_scenario(self, request, queryset):
        scenario = Scenario(name='Scenario %s' % time.ctime())
        scenario.save()
        for action in queryset:
            action.scenario = scenario
            action.save()
        return redirect('admin:replay_scenario_change', scenario.id)

    create_scenario.short_description = 'Create scenario from actions'

    def create_validators(self, request, queryset):
        title_pattern = '<title>(.*?)</title>'

        for action in queryset:
            pattern = None

            if action.status_code.startswith('3'):
                pattern = re.escape(action.content)
                action.name = '%s %s' % (action.method, action.path)
                action.save()
            else:
                match = re.search(title_pattern, action.content)

                if match:
                    title = match.group(1)
                    pattern = re.escape(title)
                    if action.name == '':
                        action.name = title
                        action.save()

            if pattern:
                Validator.objects.create(
                    action=action,
                    pattern=pattern,
                )

    create_validators.short_description = 'Create validators automatically'

    def clear_content(self, request, queryset):
        queryset.update(content='')

    clear_content.short_description = 'Clear action content'

    def get_queryset(self, request):
        queryset = super(ActionAdmin, self).get_queryset(request)
        queryset = queryset.annotate(count_validators=Count('validators__id'))
        return queryset

    def scenario_link(self, action):
        scenario = action.scenario
        url = reverse('admin:replay_scenario_change', args=(scenario.id,))
        return format_html('<a href="{}">{}</a>', url, scenario)

    scenario_link.short_description = 'Scenario Link'
    scenario_link.allow_tags = True

    def validators_count(self, instance):
        return instance.count_validators

    validators_count.admin_order_field = 'count_validators'
    validators_count.short_description = 'Validators'

    actions = (
        'create_scenario',
        'create_validators',
        'clear_content',
    )
    fields = (
        ('scenario', 'scenario_link', 'order'),
        'name',
        ('method', 'path'),
        'data',
        'files',
        'status_code',
        'content',
    )
    form = ActionAdminForm
    inlines = (
        ValidatorInline,
    )
    list_display = (
        'id',
        'name',
        'scenario',
        'order',
        'method',
        'path',
        'status_code',
        'validators_count',
    )
    ordering = (
        '-scenario',
        'order',
        'id',
    )
    readonly_fields = (
        'scenario_link',
    )
    save_on_top = True
    search_fields = (
        'name',
        'method',
        'path',
        'status_code',
    )

admin.site.register(Action, ActionAdmin)

class ValidatorAdmin(admin.ModelAdmin):
    def action_link(self, validator):
        action = validator.action
        url = reverse('admin:replay_action_change', args=(action.id,))
        return format_html('<a href="{}">{}</a>', url, action)

    fields = (
        'action',
        'action_link',
        'order',
        'pattern',
    )
    list_display = (
        'id',
        'action',
        'order',
        '__str__',
    )
    ordering = (
        '-action',
        'order',
        'id',
    )
    readonly_fields = (
        'action_link',
    )
    save_on_top = True
    search_fields = (
        'action__name',
        'action__path',
        'pattern',
    )

admin.site.register(Validator, ValidatorAdmin)

class ActionInline(admin.StackedInline):
    def action_link(self, action):
        url = reverse('admin:replay_action_change', args=(action.id,))
        return format_html('<a href="{}">{}</a>', url, action)

    def validators(self, action):
        validators = Validator.objects.filter(action=action)
        validators = validators.order_by('order', 'id')
        iterable = ((validator.pattern,) for validator in validators)
        return format_html_join(mark_safe('<br><br>'), '&bull; {}', iterable)

    model = Action
    extra = 8
    ordering = (
        'order',
        'id',
    )
    fields = (
        'action_link',
        'order',
        'name',
        ('method', 'path'),
        'data',
        'files',
        'status_code',
        'content',
        'validators',
    )
    form = ActionAdminForm
    readonly_fields = (
        'action_link',
        'validators',
    )

class ScenarioAdminForm(forms.ModelForm):
    class Meta:
        model = Scenario
        fields = []  # Required but ignored.
        widgets = {
            'name': forms.TextInput(attrs={'size': 80}),
        }

class ScenarioAdmin(admin.ModelAdmin):
    fields = (
        'name',
        'priority',
    )
    form = ScenarioAdminForm
    inlines = (
        ActionInline,
    )
    list_display = (
        'id',
        'name',
        'priority',
    )
    ordering = (
        '-id',
    )
    save_on_top = True
    search_fields = (
        'name',
    )

admin.site.register(Scenario, ScenarioAdmin)



"""Test Scenario Management Command

"""

import base64
import json
import re
import string
import uuid

from django.core.management.base import BaseCommand, CommandError
from django.test import Client

from replay.models import Scenario, Action, Validator

def expand(text, mapping):
    template = string.Template(text)
    return template.substitute(mapping)


class Command(BaseCommand):
    can_import_settings = True

    def add_arguments(self, parser):
        parser.add_argument('scenario', nargs='*')


    def handle(self, *args, **options):
        try:
            scenarios = []
            for name in options['scenario']:
                scenario = Scenario.objects.get(name=name)
                scenarios.append(scenario)
        except Scenario.DoesNotExist:
            message = 'scenario with name %r does not exist' % name
            raise CommandError(message)
        else:
            if not scenarios:
                scenarios = Scenario.objects.all().order_by('priority', 'id')

        self.client = Client()  # pylint: disable=attribute-defined-outside-init
        state = {}

        for scenario in scenarios:
            self.run(scenario, state)


    def run(self, scenario, state):
        actions = Action.objects.filter(scenario=scenario).order_by('order', 'id')
        errors = 0

        self.stdout.write(repr(scenario))
        for action in actions:
            print(' ' * 4 + repr(action))
            status_code, content = self.request(action, state)

            if status_code != action.status_code:
                message = 'FAIL %r status code: %s expected: %s'
                data = action, status_code, action.status_code
                self.stdout.write(message % data)
                self.stdout.write('FAIL STATE')
                for key, value in state.items():
                    self.stdout.write('FAIL %s = %s' % (key, value))

                errors += 1
                break

            validators = Validator.objects.filter(action=action)
            validators = validators.order_by('order', 'id')

            if validators:
                if isinstance(content, bytes):
                    content = content.decode('utf-8')

            for validator in validators:
                pattern = expand(validator.pattern, state)
                match = re.search(pattern, content)

                if match:
                    self.stdout.write(' ' * 8 + repr(validator))
                    state.update(match.groupdict())
                else:
                    self.stdout.write(' ' * 8 + 'FAIL')
                    self.stdout.write(' ' * 8 + 'Pattern: ' + pattern)
                    self.stdout.write(' ' * 8 + 'Content: ' + content)
                    errors += 1

        if errors:
            raise CommandError('%s errors' % errors)


    def request(self, action, state):
        func = getattr(self.client, action.method.lower())
        uuid_bytes = uuid.uuid4().bytes
        uuid_base64 = base64.urlsafe_b64encode(uuid_bytes)
        uuid_clean = uuid_base64.strip(b'=').decode()
        state['__uuid'] = uuid_clean
        data = json.loads(expand(action.data, state))
        files = json.loads(expand(action.files, state))
        path = expand(action.path, state)

        for key, value in files.items():
            data[key] = open(value, 'rb')

        response = func(path, data)
        status_code = str(response.status_code)
        redirect = '300' <= status_code < '400'
        content = response.url if redirect else response.content

        return status_code, content



# -*- coding: utf-8 -*-
# Generated by Django 1.11.17 on 2019-01-04 05:11
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('replay', '0006_auto_20170225_0018'),
    ]

    operations = [
        migrations.AlterField(
            model_name='action',
            name='data',
            field=models.TextField(help_text='HTTP parameters in JSON format. Supports string templating. Example: $name or $(name).'),
        ),
        migrations.AlterField(
            model_name='action',
            name='files',
            field=models.TextField(help_text='Uploaded files in JSON format. Key and value pairs correspond to filename and file-path pairs. Supports string templating. Example: $name and $(name).'),
        ),
        migrations.AlterField(
            model_name='action',
            name='method',
            field=models.CharField(choices=[('GET', 'GET'), ('POST', 'POST')], max_length=8),
        ),
        migrations.AlterField(
            model_name='action',
            name='name',
            field=models.TextField(blank=True, default=''),
        ),
        migrations.AlterField(
            model_name='action',
            name='order',
            field=models.FloatField(default=1.0, help_text='Actions are run in ascending order.'),
        ),
        migrations.AlterField(
            model_name='scenario',
            name='priority',
            field=models.FloatField(default=1.0, help_text='Scenarios are run in ascending priority.'),
        ),
        migrations.AlterField(
            model_name='validator',
            name='order',
            field=models.FloatField(default=1.0, help_text='Validators are run in ascending order.'),
        ),
        migrations.AlterField(
            model_name='validator',
            name='pattern',
            field=models.TextField(help_text='Matched against HTTP response. Supports regular expressions. Symbolic group names are stored and may be used for string templating. Example: (?P&lt;name&gt;pattern)'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from collections import defaultdict

from django.db import models, migrations

def forwards_func(apps, schema_editor):
    db_alias = schema_editor.connection.alias

    Scenario = apps.get_model('replay', 'Scenario')
    Action = apps.get_model('replay', 'Action')
    Step = apps.get_model('replay', 'Step')

    mapping = defaultdict(list)

    for step in Step.objects.using(db_alias).all():
        mapping[step.action_id].append(step)

    for action_id in list(mapping):
        steps = mapping[action_id]

        while len(steps) > 1:
            step = steps.pop()
            step_action = step.action
            action = Action.objects.using(db_alias).create(
                name=step_action.name,
                method=step_action.method,
                path=step_action.path,
                data=step_action.data,
                files=step_action.files,
                status_code=step_action.status_code,
                content=step_action.content,
            )
            step.action = action
            step.save()

    mapping.clear()

    for step in Step.objects.using(db_alias).all():
        mapping[step.action_id].append(step)

    for action_id, (step,) in mapping.items():
        action = step.action
        action.scenario = step.scenario
        action.order = step.order
        action.save()


class Migration(migrations.Migration):

    dependencies = [
        ('replay', '0002_action_files'),
    ]

    operations = [
        migrations.AddField(
            model_name='action',
            name='order',
            field=models.FloatField(default=1.0),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='action',
            name='scenario',
            field=models.ForeignKey(default=None, blank=True, on_delete=models.CASCADE, to='replay.Scenario', null=True),
            preserve_default=True,
        ),
        migrations.RunPython(
            forwards_func,
        ),
    ]



"""Record Management Command

TODO

* Add support for "runserver" options.

"""

import itertools

from django.core.management import call_command
from django.core.management.base import BaseCommand

class Command(BaseCommand):
    can_import_settings = True

    def handle(self, *args, **options):
        # pylint: disable=import-outside-toplevel
        from django.conf import settings
        iterator = itertools.chain(
            ('replay.middleware.RecorderMiddleware',),
            settings.MIDDLEWARE,
        )
        settings.MIDDLEWARE = tuple(iterator)
        call_command('runserver', *args, **options)
